// namespace:type/location
type NamespacedId = `${string}:${"img" | "audio" | "sprite"}/${string}`;

/**
 * Resources are stored in namespaces, and can be accessed by their namespaced id.
 * Sprites are located in blob storage as a single png file.
 * Audio is located in blob storage as a single mp3 file.
 *
 * Custom resources can be loaded via the public API, however they will not be loaded on other clients.
 * Ideally, engine and car definitions should be stored in the resource manager so that custom cars can be created.
 */
export class ResourceManager {
  private resources: Map<string, unknown> = new Map();
  private statuses: Map<string, Promise<boolean>> = new Map();

  get<T>(name: NamespacedId): T {
    if (!this.resources.has(name)) {
      throw new Error(`Resource ${name} not found`);
    }
    return this.resources.get(name) as T;
  }

  set(
    name: NamespacedId,
    value: unknown,
  ) {
    const identifier = parseNamespacedId(name);
    if (typeof (value as EventSource).addEventListener === "function") {
      if (value instanceof Image) {
        // During development, we can use the local file system
        value.src =
          `/blobs/${identifier.namespace}/${identifier.type}/${identifier.name}${
            extensionByType(identifier.type)
          }`;
        console.log(value.src);
      }
      this.statuses.set(
        name,
        new Promise((resolve) => {
          const onload = () => {
            this.resources.set(name, value);
            resolve(true);
            (value as EventSource).removeEventListener("load", onload);
          };
          (value as EventSource).addEventListener("load", onload);
        }),
      );
    } else {
      console.warn("Resource added was not a loadable resource");
    }
    this.resources.set(name, value);
  }

  delete(name: string) {
    this.resources.delete(name);
  }

  ready() {
    return Promise.all(Array.from(this.statuses.values()));
  }
}

type ResourceType = "img" | "audio" | "sprite";

function extensionByType(type: ResourceType) {
  switch (type) {
    case "img":
      return ".png";
    case "audio":
      return ".mp3";
    case "sprite":
      return ".png";
  }
}

type NamespaceIdentifier = {
  namespace: string;
  type: ResourceType;
  name: string;
};

function parseNamespacedId(id: NamespacedId): NamespaceIdentifier {
  const [namespace, location] = id.split(":");
  const [type, ...name] = location.split("/");
  return { namespace, type: type as ResourceType, name: name.join("/") };
}
